@@ -3,7 +3,6 @@ package internal
 import (
 	"fmt"
 	"math"
-	"regexp"
 	"runtime/debug"
 	"sync"
 	"time"
@@ -12,6 +11,7 @@ import (
 	log "github.com/sirupsen/logrus"
 	"github.com/walles/moor/v2/internal/linemetadata"
 	"github.com/walles/moor/v2/internal/reader"
+	"github.com/walles/moor/v2/internal/search"
 	"github.com/walles/moor/v2/internal/textstyles"
 	"github.com/walles/moor/v2/twin"
 )
@@ -64,12 +64,15 @@ type Pager struct {
 	// mode is better?
 	mode PagerMode
 
-	searchString  string
-	searchPattern *regexp.Regexp
-	filterPattern *regexp.Regexp
+	search search.Search
+
+	// This should never be null while paging. Configured in NewPager().
+	searchHistory *SearchHistory
+
+	filter search.Search
 
 	// We used to have a "Following" field here. If you want to follow, set
-	// TargetLineNumber to LineNumberMax() instead, see below.
+	// TargetLineNumber to linemetadata.IndexMax() instead, see below.
 
 	isShowingHelp bool
 	preHelpState  *_PreHelpState
@@ -115,7 +118,12 @@ type Pager struct {
 
 	WithTerminalFg bool // If true, don't set linePrefix
 
-	// Length of the longest line displayed. This is used for limiting scrolling to the right.
+	// If false, don't highlight lines with search hits (but still highlight the
+	// actual hits)
+	WithSearchHitLineBackground bool
+
+	// Length of the longest line displayed. This is used for limiting scrolling
+	// to the right.
 	longestLineLength int
 
 	// Bookmarks that you can come back to.
@@ -179,6 +187,7 @@ Searching
 * Type / to start searching, then type what you want to find
 * Type ? to search backwards, then type what you want to find
 * Type RETURN to stop searching, or ESC to skip back to where the search started
+* Press up / down arrows while searching to access search history
 * Find next by typing 'n' (for "next")
 * Find previous by typing SHIFT-N or 'p' (for "previous")
 * Search is case sensitive if it contains any UPPER CASE CHARACTERS
@@ -213,27 +222,31 @@ func NewPager(readers ...*reader.ReaderImpl) *Pager {
 	}
 
 	pager := Pager{
-		readers:          readers,
-		currentReader:    0,
-		readerSwitched:   make(chan struct{}, 1),
-		quit:             false,
-		ShowLineNumbers:  true, // Constant throghout the lifetime of the pager
-		showLineNumbers:  true, // Will be updated over time
-		ShowStatusBar:    true,
-		DeInit:           true,
-		SideScrollAmount: 16,
-		TabSize:          8, // This is what less defaults to
-		ScrollLeftHint:   textstyles.CellWithMetadata{Rune: '<', Style: twin.StyleDefault.WithAttr(twin.AttrReverse)},
-		ScrollRightHint:  textstyles.CellWithMetadata{Rune: '>', Style: twin.StyleDefault.WithAttr(twin.AttrReverse)},
-		scrollPosition:   newScrollPosition(name),
+		readers:                     readers,
+		currentReader:               0,
+		readerSwitched:              make(chan struct{}, 1),
+		quit:                        false,
+		ShowLineNumbers:             true, // Constant throghout the lifetime of the pager
+		showLineNumbers:             true, // Will be updated over time
+		ShowStatusBar:               true,
+		DeInit:                      true,
+		SideScrollAmount:            16,
+		TabSize:                     8, // This is what less defaults to
+		ScrollLeftHint:              textstyles.CellWithMetadata{Rune: '<', Style: twin.StyleDefault.WithAttr(twin.AttrReverse)},
+		ScrollRightHint:             textstyles.CellWithMetadata{Rune: '>', Style: twin.StyleDefault.WithAttr(twin.AttrReverse)},
+		scrollPosition:              newScrollPosition(name),
+		WithSearchHitLineBackground: true,
 	}
 
 	pager.mode = PagerModeViewing{pager: &pager}
 	pager.filteringReader = FilteringReader{
 		BackingReader: readers[0], // Always start with the first reader
-		FilterPattern: &pager.filterPattern,
+		Filter:        &pager.filter,
 	}
 
+	searchHistory := BootSearchHistory("")
+	pager.searchHistory = &searchHistory
+
 	return &pager
 }
 
@@ -272,6 +285,37 @@ func (p *Pager) getLineNumberPrefixLength(lineNumber linemetadata.Number) int {
 	return length
 }
 
+// Single quoted parts of the help text will be highlighted relative to the
+// status bar style.
+func renderHelpText(help string) []twin.StyledRune {
+	var result []twin.StyledRune
+
+	shortcutHighlight := twin.AttrBold
+	if statusbarStyle.HasAttr(shortcutHighlight) {
+		shortcutHighlight = twin.AttrUnderline
+	}
+	if statusbarStyle.HasAttr(shortcutHighlight) {
+		shortcutHighlight = twin.AttrReverse
+	}
+
+	style := statusbarStyle
+	for _, token := range help {
+		if token == '\'' {
+			// Highlight things within single quotes
+			if style == statusbarStyle {
+				style = statusbarStyle.WithAttr(shortcutHighlight)
+			} else {
+				style = statusbarStyle
+			}
+			continue
+		}
+
+		result = append(result, twin.NewStyledRune(token, style))
+	}
+
+	return result
+}
+
 // Draw the footer string at the bottom using the status bar style.
 //
 // Single quoted parts of the help text will be bolded.
@@ -289,25 +333,8 @@ func (p *Pager) setFooter(footer string, help string) {
 	}
 
 	// Help text, highlight keyboard shortcuts
-	highlightAttr := twin.AttrBold
-	if statusbarStyle.HasAttr(highlightAttr) {
-		highlightAttr = twin.AttrUnderline
-	}
-	if statusbarStyle.HasAttr(highlightAttr) {
-		highlightAttr = twin.AttrReverse
-	}
-	style := statusbarStyle
-	for _, token := range help {
-		if token == '\'' {
-			// Highlight things within single quotes
-			if style == statusbarStyle {
-				style = statusbarStyle.WithAttr(highlightAttr)
-			} else {
-				style = statusbarStyle
-			}
-			continue
-		}
-		pos += p.screen.SetCell(pos, height-1, twin.NewStyledRune(token, style))
+	for _, cell := range renderHelpText(help) {
+		pos += p.screen.SetCell(pos, height-1, cell)
 	}
 
 	for pos < width {
@@ -392,9 +419,8 @@ func (p *Pager) setTargetLine(targetLine *linemetadata.Index) {
 		return
 	}
 
-	// The value 1000 here is supposed to be larger than any possible screen
-	// height, to give us some lookahead and to avoid fetching too few lines.
-	targetValue := targetLine.Index() + 1000
+	// Set the target with some lookahead to avoid fetching too few lines.
+	targetValue := targetLine.Index() + reader.DEFAULT_PAUSE_AFTER_LINES/2
 	if targetValue < targetLine.Index() {
 		// Overflow detected, clip to max int
 		targetValue = math.MaxInt
@@ -428,7 +454,7 @@ func (p *Pager) StartPaging(screen twin.Screen, chromaStyle *chroma.Style, chrom
 		textstyles.TabSize = p.TabSize
 	}
 	consumeLessTermcapEnvs(screen.TerminalBackground(), chromaStyle, chromaFormatter)
-	styleUI(screen.TerminalBackground(), chromaStyle, chromaFormatter, p.StatusBarStyle, p.WithTerminalFg)
+	styleUI(screen.TerminalBackground(), chromaStyle, chromaFormatter, p.StatusBarStyle, p.WithTerminalFg, p.WithSearchHitLineBackground)
 
 	p.screen = screen
 	p.mode = PagerModeViewing{pager: p}
@@ -462,7 +488,7 @@ func (p *Pager) StartPaging(screen twin.Screen, chromaStyle *chroma.Style, chrom
 			select {
 			case <-p.readerSwitched:
 				// A different reader is now active
-				p.filterPattern = nil
+				p.filter = search.Search{}
 
 				p.readerLock.Lock()
 				r = p.readers[p.currentReader]
@@ -494,7 +520,7 @@ func (p *Pager) StartPaging(screen twin.Screen, chromaStyle *chroma.Style, chrom
 
 			case <-spinnerTicker.C:
 				currentSpinnerFrame := spinnerFrames[spinnerIndex]
-				if r.Done.Load() {
+				if r.ReadingDone.Load() {
 					// We're done, clear the spinner
 					currentSpinnerFrame = ""
 				}
@@ -539,7 +565,7 @@ func (p *Pager) StartPaging(screen twin.Screen, chromaStyle *chroma.Style, chrom
 			//
 			// Also, we only do this if we have exactly one reader, because
 			// that's what less does.
-			if len(p.readers) == 1 && p.QuitIfOneScreen && !p.isShowingHelp && r.Done.Load() && r.HighlightingDone.Load() {
+			if len(p.readers) == 1 && p.QuitIfOneScreen && !p.isShowingHelp && r.ReadingDone.Load() && r.HighlightingDone.Load() {
 				if p.fitsOnOneScreen() {
 					// Ref:
 					// https://github.com/walles/moor/issues/113#issuecomment-1368294132
@@ -593,7 +619,10 @@ func (p *Pager) StartPaging(screen twin.Screen, chromaStyle *chroma.Style, chrom
 			return
 
 		case eventMoreLinesAvailable:
-			if p.TargetLine != nil {
+			// Without the isViewing() check, following will continue while
+			// searching, and I prefer it to stop so people can see what they
+			// are searching in.
+			if p.isViewing() && p.TargetLine != nil {
 				// The user wants to scroll down to a specific line number
 				if linemetadata.IndexFromLength(p.Reader().GetLineCount()).IsBefore(*p.TargetLine) {
 					// Not there yet, keep scrolling
@@ -606,8 +635,17 @@ func (p *Pager) StartPaging(screen twin.Screen, chromaStyle *chroma.Style, chrom
 			}
 
 		case eventMaybeDone:
-			// Do nothing. We got this just so that we'll do the QuitIfOneScreen
-			// check (above) as soon as highlighting is done.
+			// Man pages come pre-formatted for the screen width, and line
+			// numbers will mess that up. So we disable line numbers if we
+			// detect a man page by its contents.
+			//
+			// See also noLineNumbersDefault() where we use environment
+			// variables to try to detect man paging.
+			if p.haveLoadedManPage() && len(p.readers) == 1 {
+				p.ShowLineNumbers = false
+				p.showLineNumbers = false
+				log.Info("man page detected by contents, disabling line numbers")
+			}
 
 		case eventSpinnerUpdate:
 			spinner = event.spinner
@@ -687,7 +725,7 @@ func (p *Pager) fitsOnOneScreen() bool {
 
 	lines := reader.GetLines(linemetadata.Index{}, reader.GetLineCount())
 	for _, line := range lines.Lines {
-		rendered := line.HighlightedTokens(twin.StyleDefault, twin.StyleDefault, nil, nil).StyledRunes
+		rendered := line.HighlightedTokens(twin.StyleDefault, twin.StyleDefault, search.Search{}).StyledRunes
 		if len(rendered) > width {
 			// This line is too long to fit on one screen line, no fit
 			return false
